from .basic_api_errors import BasicAPIErrorTest
from .basic_api_test import BasicAPITest
from .test_auth_api import AuthAPITest
from .test_cache import TestCache
from .test_locale_api import LocaleAPITest
from .test_perms_api import PermsAPITest
